<?php

namespace Nwb\Satellites;

use CURLFile;
use Exception;
use InvalidArgumentException;

class SatelliteClient
{
    private string $apiKey;
    private string $user;
    private string $password;
    private HttpClient $httpClient;

    public function __construct(string $user = null, string $password = null, string $apiKey = null)
    {
        $this->user = ($user ?? '');
        if (empty($this->user)) $this->user = getenv('SAT_USER');
        if (empty($this->user)) $this->user = getenv('SAT_LOGIN');

        $this->password = ($password ?? '');
        if (empty($this->password)) $this->password = getenv('SAT_PWD');
        if (empty($this->password)) $this->password = getenv('SAT_MDP');

        $this->apiKey = ($apiKey ?? '');
        if (empty($this->apiKey)) $this->apiKey = getenv('SAT_APIKEY');
        if (empty($this->apiKey)) $this->apiKey = getenv('SAT_API_KEY');

        if (!$this->user) {
            throw new Exception('User is missing');
        }

        if (!$this->password) {
            throw new Exception('Password is missing');
        }

        if (!$this->apiKey) {
            throw new Exception('API key is missing');
        }

        $this->httpClient = new HttpClient();
    }

    /**
     * Uploads a file to the Satellites platform.
     *
     * @param File $file The file to upload.
     * @throws SatelliteException If an error occurs during the upload process.
     */
    /**
     * This class represents a client for interacting with the Satellites API.
     * It provides methods for uploading, deleting and retrieving files from the Satellites platform.
     */
    public function uploadFile(File $file): void
    {
        $options = [
            CURLOPT_POST => 1,
            CURLOPT_POSTFIELDS => [
                'SAI_UPLOAD' => new CURLFile($file->getPath()),
            ],
            CURLOPT_RETURNTRANSFER => 1,
            CURLOPT_HTTPHEADER => [
                'LOGIN-UTILISATEUR: ' . $this->user,
                'MDP-UTILISATEUR: ' . $this->password,
                'HASH: ' . $file->getHash(),
                'TAILLE: ' . $file->getSize(),
                'TYPE-ACCES: ' . $file->getAccessType()->value,
            ]
        ];

        $numTries = 0;

        // Si le satellite ne return pas le hash du fichier, on retente
        while (true) {
            try {
                $satelliteMapper = new SatelliteMapper();
                $satellite = $satelliteMapper->chooseSatelliteForUpload($file->getSize());

                $url = 'https://' . $satellite->getDomain() . '/FR/AddF-2.php';

                $response = $this->httpClient->request($url, $options);

                if ($error = $this->parseErrorMessage($response)) {
                    throw new SatelliteException($error . ' URL : ' . $url);
                }

                $response = explode('[;]', $response);

                if ($response[2] !== $file->getHash()) {
                    throw new SatelliteException($response[0] . ' URL: ' . $url);
                }

                break;
            } catch (Exception $e) {
                $numTries++;

                if ($numTries === 3) {
                    throw $e;
                }
            }
        }

        $file->setSatelliteId($satellite->getId());
    }

    public function deleteFile(File $file): void
    {
        $satellite = new Satellite($file->getSatelliteId());

        $url = 'https://' . $satellite->getDomain() . '/FR/DelF-2.php';

        $options = [
            CURLOPT_RETURNTRANSFER => 1,
            CURLOPT_HTTPHEADER => [
                'LOGIN-UTILISATEUR: ' . $this->user,
                'MDP-UTILISATEUR: ' . $this->password,
                'H: ' . $file->getHash(),
                'T: ' . $file->getSize(),
                'L: ' . $this->user,
                'SUPPRIME-LIEN-BDD: 1',
            ]
        ];

        $response = $this->httpClient->request($url, $options);

        if ($error = $this->parseErrorMessage($response)) {
            throw new SatelliteException($error);
        }

        $file->setSatelliteId(null);
    }

    private function parseErrorMessage(string | null $satelliteResponse): string
    {
        if (!$satelliteResponse) return 'Empty response';
        $parts = explode(':', $satelliteResponse);

        if (trim($parts[0]) === '-1') {
            return trim($parts[1]);
        }

        return '';
    }

    public function getFileUrlParams(File $file): string
    {
        $url = 'H=' . $file->getHash();
        $url .= '&T=' . $file->getSize();

        if ($file->getAccessType() === AccessType::PRIVATE) {
            $chaine = $file->getHash() . $file->getSize() . $this->user . date('Ymd');
            $htc = strtoupper(hash_hmac('sha1', $chaine, $this->apiKey));

            $url .= '&L=' . $this->user;
            $url .= '&HTC=' . $htc;
        }

        return $url;
    }

    public function getFileUrl(File $file): string
    {
        $satellite = new Satellite($file->getSatelliteId());

        return 'https://' . $satellite->getDomain() . '/FR/GetF.php?' . $this->getFileUrlParams($file);
    }

    public function downloadFileTo(File $file, string $destination): void
    {
        $url = $this->getFileUrl($file);

        // Vérification que le répertoire de destination existe
        $dir = dirname($destination);
        if (!is_dir($dir)) throw new InvalidArgumentException('Invalid destination directory does not exist : ' . $dir);

        $fileHandle  = fopen($destination, 'w');
        if (!$fileHandle) throw new InvalidArgumentException('Invalid destination file');

        $options = [
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_FILE => $fileHandle
        ];

        try {
            $this->httpClient->request($url, $options);
        } catch (SatelliteException $e) {
            $satellite = new Satellite($file->getSatelliteId());
            $response = $this->httpClient->request($satellite->getFallbackUrl($url), $options);
            $error = $this->parseErrorMessage($response);
            throw new SatelliteException($error . ' URL : ' . $url);
        }
    }

    public function downloadFileAsString(File $file): string
    {
        $url = $this->getFileUrl($file);

        $options = [
            CURLOPT_RETURNTRANSFER => true,
        ];

        try {
            $response = $this->httpClient->request($url, $options);
        } catch (SatelliteException $e) {
            $satellite = new Satellite($file->getSatelliteId());

            $response = $this->httpClient->request($satellite->getFallbackUrl($url), $options);
        }

        if ($error = $this->parseErrorMessage($response)) {
            throw new SatelliteException($error . ' URL : ' . $url);
        }

        return $response;
    }

    private function generateSecretHash(File $file, int $height, int $quality, File $watermark = null): string
    {
        $hash = $this->user . $file->getSize() . $file->getHash() . $height . $quality;

        if ($watermark !== null) {
            $hash .= $watermark->getHashSize();
        }

        $hash .= date('Ymd');

        return strtoupper(hash_hmac('sha1', $hash, $this->apiKey));
    }

    /**
     * $attributes
     *     width => int,
     *     height => int,
     *     quality => int 0 à 100,
     *     orientation => numéro orientation exif ou 'auto' pour détection par les sat
     *     crop => string 'auto' (Homothétique Étendu Centré)
     */
    public function getThumbnailUrlParams(
        File $photo,
        array $attributes = [],
        File $watermark = null
    ): string {
        $height = $attributes['height'] ?? null;
        $width = $attributes['width'] ?? null;
        $quality = $attributes['quality'] ?? 80;
        $orientation = $attributes['orientation'] ?? null;
        $crop = isset($attributes['crop']) && $attributes['crop'] === 'auto';

        $height = !$height && !$width ? 500 : $height;

        $params = 'H=' . $photo->getHash();
        $params .= '&T=' . $photo->getSize();
        $params .= '&L=' . $this->user;

        if ($width) {
            $params .= '&X=' . $width;
        }

        if ($height) {
            $params .= '&Y=' . $height;
        }

        $params .= '&Q=' . $quality;

        if (is_int($orientation)) {
            $params .= '&R=1&OI=' . $orientation;
        } elseif ($orientation === 'auto') {
            $params .= '&R=1';
        }

        if ($crop) {
            $params .= '&HEC=1';
        }

        if ($watermark !== null) {
            $params .= '&HTW=' . $watermark->getHashSize();
        }

        $params .= '&C=' . $this->generateSecretHash($photo, $height, $quality, $watermark);

        return $params;
    }

    public function getThumbnailUrl(
        File $photo,
        array $attributes = [],
        File $watermark = null
    ): string {
        $satellite = new Satellite($photo->getSatelliteId());

        return 'https://' . $satellite->getDomain() . '/FR/ImgMini.php?' . $this->getThumbnailUrlParams($photo, $attributes, $watermark);
    }

    public function downloadThumbnail(File $photo, array $attributes = [], File $watermark = null): string
    {
        $url = $this->getThumbnailUrl($photo, $attributes, $watermark);

        $options = [
            CURLOPT_RETURNTRANSFER => 1,
        ];

        try {
            $response = $this->httpClient->request($url, $options);
        } catch (SatelliteException $e) {
            $satellite = new Satellite($photo->getSatelliteId());

            $response = $this->httpClient->request($satellite->getFallbackUrl($url), $options);
        }

        if ($error = $this->parseErrorMessage($response)) {
            throw new SatelliteException($error . ' URL : ' . $url);
        }

        return $response;
    }

    public function getMetaImage(File $file, string $type = 'all-exiftool-json')
    {
        if (!in_array($type, ['orientation', 'all', 'timeshoot', 'all-exiftool-json'])) {
            throw new InvalidArgumentException("Le type [$type] est invalide");
        }

        $satellite = new Satellite($file->getSatelliteId());

        $url = 'https://' . $satellite->getDomain() . '/FR/Get-meta-image.php';
        $url .= '?H=' . $file->getHash();
        $url .= '&T=' . $file->getSize();
        $url .= '&M=' . $type;

        $options = [
            CURLOPT_RETURNTRANSFER => 1,
        ];

        $data = $this->httpClient->request($url, $options);

        if ($type === 'all-exiftool-json') {
            return json_decode($data, true);
        }

        return $data;
    }

    public function giveFile(File $file, string $recipient): void
    {
        $satellite = new Satellite($file->getSatelliteId());

        $url = 'https://' . $satellite->getDomain() . '/FR/DonneF-2.php';

        $options = [
            CURLOPT_RETURNTRANSFER => 1,
            CURLOPT_HTTPHEADER => [
                'LOGIN-UTILISATEUR: ' . $this->user,
                'MDP-UTILISATEUR: ' . $this->password,
                'H: ' . $file->getHash(),
                'T: ' . $file->getSize(),
                'L: ' . $this->user,
                'D: ' . $recipient,
                'SUPPRIME-LIEN-BDD: 1',
            ]
        ];

        $response = $this->httpClient->request($url, $options);

        if ($error = $this->parseErrorMessage($response)) {
            throw new SatelliteException($error);
        }
    }

    public function existeSelonCRC(int $satelliteId, File $file): bool
    {
        $satellite = new Satellite($satelliteId);

        $url = 'https://' . $satellite->getDomain() . '/FR/FExiste_Selon_CRC.php';
        $url .= '?H' . $file->getHash();
        $url .= '&T=' . $file->getSize();

        $options = [
            CURLOPT_RETURNTRANSFER => 1,
        ];

        $response = $this->httpClient->request($url, $options);

        return match ($response) {
            '-1' => false,
            '1' => true,
            default => throw new SatelliteException($response),
        };
    }
}
